from enum import Enum

CHANNEL_TYPE = "sales_channel"
INTEGRATION_TYPE = "sales_channel"
CHUNK_SIZE = 50
CONTENT_TYPE_IDS = {}


class BatchRequestStatus(Enum):
    """
    Batch requests work as a state machine. These are states defined for them.
    """
    # Initial state
    initialized = "initialized"
    # Channel app reports that it received data and objects in the batch are marked so that they
    # are not reprocessed on further requests
    commit = "commit"
    # Objects are sent to the Channel and being processed on Channel side
    sent_to_remote = "sent_to_remote"
    # Objects are still being processed on the Channel
    ongoing = "ongoing"
    # Channel finished processing the batch and returned the result. Some items may be in
    # fail state
    done = "done"
    # Channel failed processing the batch and halted the process.
    fail = "fail"


class ErrorType(Enum):
    country = "country"
    city = "city"
    township = "township"
    district = "district"


class ContentType(Enum):
    """
    Omnitron model content types. Batch requests can be used for all Omnitron models. We need to
    specify content type of the model in question. Both for the batch request itself and inside the
    objects block of the batch.
    """
    # TODO check all contenttype usages and add them here
    batch_request = "batchrequest"
    product = "product"
    order = "order"
    order_item = "orderitem"
    product_price = "productprice"
    product_stock = "productstock"
    integration_action = "integrationaction"
    product_image = "productimage"
    category_tree = "categorytree"
    category_node = "categorynode"
    channel = "channel"
    product_category = "productcategory"
    attribute_set = "marketplaceattributeset"
    attribute_set_config = "marketplaceattributesetconfig"
    attribute = "marketplaceattribute"
    attribute_schema = "marketplaceattributeschema"
    attribute_value = "marketplaceattributevalue"
    attribute_value_config = "marketplaceattributevalueconfig"
    cancellation_request = "cancellationrequest"


class FailedReasonType(Enum):
    mapping = "mapping"
    remote = "remote"
    channel_app = "channel_app"


class CustomerIdentifierField(Enum):
    """
    Customer uniqueness identifier field
    """
    email = "email"
    phone_number = "phone_number"


class ResponseStatus:
    fail = "FAIL"
    success = "SUCCESS"


class IntegrationActionStatus:
    processing = "processing"
    success = "success"
    error = "error"


class ChannelConfSchemaDataTypes:
    text = 'text'
    URL = 'URL'
    email = 'email'
    date = 'date'
    datetime = 'datetime'
    bool = 'bool'
    json = 'json'
    list = 'list'
    integer = 'integer'



from typing import List

from channel_app.core import settings
from channel_app.core.data import CategoryTreeDto, ErrorReportDto, AttributeDto
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.constants import ContentType


class SetupService(object):
    def create_or_update_category_tree_and_nodes(self, is_success_log=False):
        with OmnitronIntegration(
                content_type=ContentType.category_tree.value) as omnitron_integration:
            channel_integration = ChannelIntegration()
            category_tree, report, _ = channel_integration.do_action(
                key='get_category_tree_and_nodes',
                batch_request=omnitron_integration.batch_request)

            category_tree: CategoryTreeDto
            report: ErrorReportDto

            if report and (is_success_log or not report.is_ok):
                omnitron_integration.do_action(
                    key='create_error_report',
                    objects=report)

            omnitron_integration.do_action(
                key='create_or_update_category_tree_and_nodes',
                objects=category_tree)

    def create_or_update_category_attributes(self, is_success_log=False):
        with OmnitronIntegration(
                content_type=ContentType.attribute.value) as omnitron_integration:
            channel_integration = ChannelIntegration()
            category_integration_actions = omnitron_integration.do_action(
                key='get_category_ids')

            for category_ia in category_integration_actions:
                if not category_ia.remote_id:
                    continue

                category, report, data = channel_integration.do_action(
                    key='get_category_attributes',
                    objects=category_ia,
                    batch_request=omnitron_integration.batch_request
                )
                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                category = category if category.attributes else None
                omnitron_integration.do_action(
                    key='create_or_update_category_attributes',
                    objects=(category_ia, category))

    def create_or_update_attributes(self, is_success_log=False):
        with OmnitronIntegration(
                content_type=ContentType.attribute.value) as omnitron_integration:
            channel_integration = ChannelIntegration()
            attributes, report, data = channel_integration.do_action(
                key='get_attributes',
                batch_request=omnitron_integration.batch_request
            )

            attributes: List[AttributeDto]
            reports: ErrorReportDto

            if report and (is_success_log or not report.is_ok):
                omnitron_integration.do_action(
                    key='create_error_report',
                    objects=report)

            for attribute in attributes:
                attr = omnitron_integration.do_action(
                    key="create_or_update_channel_attribute",
                    objects={"name": attribute.name,
                             "remote_id": attribute.remote_id},
                )[0]

                omnitron_integration.do_action(
                    key="get_or_create_channel_attribute_schema",
                    objects={"name": f"{settings.OMNITRON_CHANNEL_ID} {attribute.name} Schema"},
                )
                for attr_value in attribute.values:
                    omnitron_integration.do_action(
                        key="create_or_update_channel_attribute_value",
                        objects={
                            "attribute": attr.pk,
                            "label": attr_value.name,
                            "value": attr_value.remote_id,
                            "remote_id": attr_value.remote_id})

    def update_channel_conf_schema(self):
        with OmnitronIntegration(
                content_type=ContentType.channel.value) as omnitron_integration:
            channel_integration = ChannelIntegration()
            schema, _, _ = channel_integration.do_action(
                key='get_channel_conf_schema',
                batch_request=omnitron_integration.batch_request)

            omnitron_integration.do_action(key="update_channel_conf_schema",
                                           objects=schema)



from dataclasses import asdict
from typing import List, Generator, Union

from omnisdk.omnitron.models import (Customer, Address, CargoCompany, Order,
                                     BatchRequest)

from channel_app.core import settings
from channel_app.core.data import (OmnitronCreateOrderDto, OmnitronOrderDto,
                                   ChannelCreateOrderDto,
                                   ErrorReportDto,
                                   OrderBatchRequestResponseDto, CancelOrderDto)
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import ContentType
from channel_app.omnitron.exceptions import (CityException,
                                             TownshipException,
                                             DistrictException,
                                             CargoCompanyException,
                                             OrderException)


class OrderService(object):
    batch_service = ClientBatchRequest

    def fetch_and_create_order(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            get_orders = ChannelIntegration().do_action(
                key='get_orders',
                batch_request=omnitron_integration.batch_request
            )

            get_orders: Generator
            while True:
                try:
                    channel_create_order, report, _ = next(get_orders)
                except StopIteration:
                    break

                # tips
                channel_create_order: ChannelCreateOrderDto
                report: ErrorReportDto

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                self.create_order(omnitron_integration=omnitron_integration,
                                  channel_order=channel_create_order)

    def create_order(self, omnitron_integration: OmnitronIntegration,
                     channel_order: ChannelCreateOrderDto
                     ) -> Union[Order, None]:
        order = channel_order.order

        try:
            customer = omnitron_integration.do_action(
                key='get_or_create_customer',
                objects=order.customer)
        except Exception:
            return

        customer: Customer
        try:
            shipping_address_data = {
                "customer": customer,
                "address": order.shipping_address
            }
            shipping_address = omnitron_integration.do_action(
                key='get_or_create_address',
                objects=shipping_address_data
            )[0]

            billing_address_data = {
                "customer": customer,
                "address": order.billing_address
            }
            billing_address = omnitron_integration.do_action(
                key='get_or_create_address',
                objects=billing_address_data
            )[0]
        except (CityException, TownshipException, DistrictException) as exc:
            omnitron_integration.do_action(
                key='create_address_error_report',
                object=exc)
            return

        shipping_address: Address
        billing_address: Address
        try:
            cargo_company = omnitron_integration.do_action(
                key='get_cargo_company',
                objects=order.cargo_company
            )[0]
        except CargoCompanyException:  # TODO: exception
            # log
            return
        cargo_company: CargoCompany

        order_data = asdict(order)
        order_data["customer"] = customer.pk
        order_data["shipping_address"] = shipping_address.pk
        order_data["billing_address"] = billing_address.pk
        order_data["cargo_company"] = cargo_company.pk
        omnitron_order = OmnitronOrderDto(**order_data)
        create_order_dto = OmnitronCreateOrderDto(
            order=omnitron_order,
            order_item=channel_order.order_item)
        try:
            orders: List[Order] = omnitron_integration.do_action(
                key='create_order',
                objects=create_order_dto
            )
            order = orders[0]  # formatted_data
        except OrderException:
            return

        omnitron_integration.do_action(
            key='create_order_shipping_info',
            objects=order
        )
        return order

    def update_orders(self, is_sync=True, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            orders = omnitron_integration.do_action(key='get_orders')
            orders: List[Order]

            if not orders:
                return

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_orders',
                objects=orders,
                batch_request=omnitron_integration.batch_request,
                is_sync=True)

            # tips
            response_data: List[OrderBatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[Order]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_order_batch_requests',
                    objects=response_data)

    def get_order_batch_requests(self, is_success_log=False):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.order.value})

            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, report, data = ChannelIntegration().do_action(
                    key='check_orders',
                    objects=batch_request)

                # tips
                response_data: List[OrderBatchRequestResponseDto]
                report: ErrorReportDto
                data: BatchRequest

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request_data
                    omnitron_integration.do_action(
                        key='process_order_batch_requests',
                        objects=response_data)

    def fetch_and_create_cancel(self, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.order.value) as omnitron_integration:
            get_cancelled_order = ChannelIntegration().do_action(
                key='get_cancelled_orders')
            get_cancelled_order: Generator

            while True:
                try:
                    cancel_order_dto, report, _ = next(get_cancelled_order)
                except StopIteration:
                    break

                # tips
                cancel_order_dto: CancelOrderDto
                report: ErrorReportDto

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                self.create_cancel(omnitron_integration=omnitron_integration,
                                   cancel_order_dto=cancel_order_dto)

    def create_cancel(self, omnitron_integration: OmnitronIntegration,
                      cancel_order_dto: CancelOrderDto):
        success_datas: List[Order] = omnitron_integration.do_action(
            key="create_order_cancel",
            objects=cancel_order_dto)
        try:
            success_data = success_datas[0]
            return success_data
        except IndexError:
            return



import logging
from datetime import datetime

from celery import current_app as app

from channel_app.core.clients import RedisClient

logger = logging.getLogger(__name__)


class LockTask(app.Task):
    """this abstract class ensures the same tasks run only once at a time"""
    abstract = True

    def __init__(self, *args, **kwargs):
        from channel_app.core import settings
        self.TTL = getattr(settings, 'DEFAULT_TASK_LOCK_TTL', 60 * 15)
        self.redis = RedisClient()
        super(LockTask, self).__init__(*args, **kwargs)

    def generate_lock_cache_key(self, *args, **kwargs):
        args_key = [str(arg) for arg in args]
        kwargs_key = ['{}_{}'.format(k, str(v)) for k, v in
                      sorted(kwargs.items())]
        return '_'.join([self.name] + args_key + kwargs_key)

    def __call__(self, *args, **kwargs):
        """check task"""
        lock_cache_key = (self.request.headers or {}).pop('cache_key', None)
        if not lock_cache_key:
            lock_cache_key = self.generate_lock_cache_key(*args, **kwargs)

        if self.lock_acquired(lock_cache_key):
            try:
                return self.run(*args, **kwargs)
            finally:
                pass
        else:
            return f'Task {self.name} is already running..'

    def lock_acquired(self, lock_cache_key):
        lock_acquired = True
        app_inspect = self.app.control.inspect()
        active_task = app_inspect.active(safe=True)
        if not active_task:
            return lock_acquired
        for worker_name, task_list in active_task.items():
            for task in task_list:
                if (task.get("name", None) == lock_cache_key) and (
                        self.request.id != task['id']):
                    lock_acquired = False
                    break
        return lock_acquired


def split_list(lst, n):
    """
    Split a list to chunks of n
    :param lst:
    :param n:
    :return iterator to get chunks
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def request_log():
    import logging
    try:
        import http.client as http_client
    except ImportError:
        import httplib as http_client

    http_client.HTTPConnection.debuglevel = 1
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def is_updated(current, new):
    no_default = "NO_DEFAULT"
    for key, new_value in new.items():
        old_value = getattr(current, key, no_default)
        if old_value != no_default and old_value != new_value:
            return True
    return False


def lowercase_keys(obj):
    if isinstance(obj, dict):
        obj = {key.lower(): value for key, value in obj.items()}
        for key, value in obj.items():
            if isinstance(value, list):
                for idx, item in enumerate(value):
                    value[idx] = lowercase_keys(value[idx])
            obj[key] = value
    return obj



from collections import defaultdict
from typing import List

from omnisdk.omnitron.endpoints import ChannelIntegrationActionEndpoint, \
    ChannelProductEndpoint, ChannelProductPriceEndpoint, \
    ChannelProductStockEndpoint
from omnisdk.omnitron.models import IntegrationAction

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.utilities import split_list


class CreateIntegrationActions(OmnitronCommandInterface):
    endpoint = ChannelIntegrationActionEndpoint

    def get_data(self) -> list:
        obj_list = self.objects
        items = []
        for obj in obj_list:
            data = {
                "content_type": obj.content_type,
                "channel_id": self.integration.channel_id,
                "object_id": obj.pk,
                "remote_id": obj.remote_id,
                "version_date": obj.modified_date,
                "state": {}
            }
            items.append(data)
        return items

    def send(self, validated_data) -> object:
        integration_actions = []
        for item in validated_data:
            integration_action = self.endpoint(
                channel_id=self.integration.channel_id).create(item=item)

            integration_actions.append(integration_action)
        return integration_actions


class UpdateIntegrationActions(CreateIntegrationActions):
    endpoint = ChannelIntegrationActionEndpoint

    def get_data(self) -> list:
        return self.objects

    def send(self, validated_data) -> object:
        integration_actions = []
        for item in validated_data:
            item.content_type_id = item.content_type['id']
            delattr(item, "content_type")

            integration_action = self.endpoint(
                channel_id=self.integration.channel_id
            ).update(id=item.pk, item=item)

            integration_actions.append(integration_action)
        return integration_actions


class GetIntegrationActionsWithObjectId(OmnitronCommandInterface):
    endpoint = ChannelIntegrationActionEndpoint
    id_type = "object_id__in"
    CHUNK_SIZE = 50

    def get_ia_dict(self, integration_action):
        return {i.object_id: i for i in integration_action}

    def get_data(self) -> List:
        integration_action_list = []
        group_by_content_type = self.get_grup_by_content_type_pk_list()

        for ct, pk_list in group_by_content_type.items():
            for chunk_pk_list in split_list(pk_list, self.CHUNK_SIZE):
                chunk_ia = self.endpoint(
                    channel_id=self.integration.channel_id
                ).list(params={
                    "limit": len(chunk_pk_list),
                    "channel_id": self.integration.channel_id,
                    "content_type_name": ct,
                    self.id_type: ",".join([str(pk) for pk in chunk_pk_list])
                })
                integration_action_list.extend(chunk_ia)

        ia_dict = self.get_ia_dict(integration_action_list)
        self.update_objects(ia_dict)
        return self.objects

    def update_objects(self, ia_dict):
        for obj in self.objects:
            obj.integration_action = ia_dict[obj.pk]

    def get_grup_by_content_type_pk_list(self):
        group_by_content_type = defaultdict(list)
        for obj in self.objects:
            group_by_content_type[obj.content_type].append(obj.pk)
        return group_by_content_type


class GetIntegrationActionsWithRemoteId(GetIntegrationActionsWithObjectId):
    endpoint = ChannelIntegrationActionEndpoint
    id_type = "remote_id__in"

    def get_grup_by_content_type_pk_list(self):
        group_by_content_type = defaultdict(list)
        for obj in self.objects:
            group_by_content_type[obj.content_type].append(obj.remote_id)
        return group_by_content_type

    def get_ia_dict(self, ia):
        return {i.remote_id: i for i in ia}

    def update_objects(self, ia_dict):
        for obj in self.objects:
            obj.integration_action = ia_dict[obj.remote_id]


class GetIntegrationActions(OmnitronCommandInterface):
    endpoint = ChannelIntegrationActionEndpoint
    CHUNK_SIZE = 10

    def get_data(self) -> dict:
        if "limit" not in self.objects:
            self.objects.update({"limit": self.CHUNK_SIZE})
        return self.objects

    def run(self) -> List[IntegrationAction]:
        integration_action_endpoint = ChannelIntegrationActionEndpoint(
            channel_id=self.integration.channel_id)
        self.objects.update({
            "channel_id": self.integration.channel_id,
            "sort": "id"
        })
        integration_action_list = integration_action_endpoint.list(
            params=self.objects)
        for batch in integration_action_endpoint.iterator:
            if not batch:
                break
            integration_action_list.extend(batch)
        return integration_action_list


class GetObjectsFromIntegrationAction(OmnitronCommandInterface):
    endpoint = {
        "product": ChannelProductEndpoint,
        "productprice": ChannelProductPriceEndpoint,
        "productstock": ChannelProductStockEndpoint
    }
    CHUNK_SIZE = 100

    def get_data(self) -> List[IntegrationAction]:
        """

        :return: Added content type object in integration action objects
            Example: if content type product, -> integration_action.product
                     if content type productprice, -> integration_action.productprice
        """
        integration_actions = self.objects
        group = self.group_by_content_type(integration_actions)
        grouped_integration_actions = self.get_content_objects(group)
        return self.extract_group(grouped_integration_actions)

    def group_by_content_type(self, integration_actions):
        """

        :param integration_actions:
        :return:
        {
        "product":[integration1,integration2,...],
        "productprice":[integration3,integration4,...]
        }
        """
        group = defaultdict(list)
        for integration_action in integration_actions:
            model = integration_action.content_type.get("model")
            group[model].append(integration_action)
        return group

    def get_content_objects(self, grouped_integration_actions: dict):
        for content_type, integration_actions in grouped_integration_actions.items():
            endpoint = self.endpoint.get(content_type)
            if endpoint:
                id_list = [str(ia.object_id) for ia in integration_actions]
                end_point = endpoint(channel_id=self.integration.channel_id)
                objects_list = []
                for chunk_id_list in split_list(id_list, self.CHUNK_SIZE):
                    objects = end_point.list(
                        params={"pk__in": ",".join(chunk_id_list),
                                "limit": len(chunk_id_list)})
                    objects_list.extend(objects)
                objects_dict = {s.pk: s for s in objects_list}
                for integration_action in integration_actions:
                    setattr(integration_action, content_type,
                            objects_dict.get(integration_action.object_id))

        return grouped_integration_actions

    def extract_group(self, group: dict):
        object_list = []
        for key, value_list in group.items():
            object_list.extend(value_list)
        return object_list



import requests

from channel_app.channel.commands.orders.orders import (
    GetOrders, CheckOrders, SendUpdatedOrders, GetCancelledOrders,
    GetUpdatedOrderItems)
from channel_app.channel.commands.product_images import (
    SendUpdatedImages, SendInsertedImages, CheckImages)
from channel_app.channel.commands.product_prices import (
    CheckPrices, SendInsertedPrices, SendUpdatedPrices)
from channel_app.channel.commands.product_stocks import (
    SendUpdatedStocks, CheckStocks, SendInsertedStocks)
from channel_app.channel.commands.products import (
    SendInsertedProducts, CheckProducts, CheckDeletedProducts,
    SendUpdatedProducts, SendDeletedProducts)
from channel_app.channel.commands.setup import (
    GetCategoryTreeAndNodes, GetCategoryAttributes, GetChannelConfSchema,
    GetAttributes)
from channel_app.core.integration import BaseIntegration


class ChannelIntegration(BaseIntegration):
    """
    Communicates with the Channel Api services through the commands defined.

    If an Api Client class is developed, initialization and deletion should be handled in
    ChannelIntegration class so that commands have easier access to the api object.
    """
    _sent_data = {}
    actions = {
        "send_inserted_products": SendInsertedProducts,
        "send_updated_products": SendUpdatedProducts,
        "send_deleted_products": SendDeletedProducts,
        "check_products": CheckProducts,
        "check_deleted_products": CheckDeletedProducts,
        "send_updated_stocks": SendUpdatedStocks,
        "send_inserted_stocks": SendInsertedStocks,
        "send_updated_prices": SendUpdatedPrices,
        "send_inserted_prices": SendInsertedPrices,
        "send_updated_images": SendUpdatedImages,
        "send_inserted_images": SendInsertedImages,
        "check_stocks": CheckStocks,
        "check_prices": CheckPrices,
        "check_images": CheckImages,
        "get_category_tree_and_nodes": GetCategoryTreeAndNodes,
        "get_channel_conf_schema": GetChannelConfSchema,
        "get_category_attributes": GetCategoryAttributes,
        "get_attributes": GetAttributes,
        "get_orders": GetOrders,
        "get_updated_order_items": GetUpdatedOrderItems,
        "send_updated_orders": SendUpdatedOrders,
        "check_orders": CheckOrders,
        "get_cancelled_orders": GetCancelledOrders
    }

    def __init__(self):
        from channel_app.core import settings
        self.channel_id = settings.OMNITRON_CHANNEL_ID
        self.catalog_id = settings.OMNITRON_CATALOG_ID

    def create_session(self):
        from channel_app.core import settings

        session = requests.Session()
        connections = self.channel.conf.get(
            'connection_pool_count', settings.DEFAULT_CONNECTION_POOL_COUNT)
        max_size = self.channel.conf.get(
            'connection_pool_max_size', settings.DEFAULT_CONNECTION_POOL_MAX_SIZE)
        retry = self.channel.conf.get(
            'connection_pool_retry', settings.DEFAULT_CONNECTION_POOL_RETRY)

        adapter = requests.adapters.HTTPAdapter(pool_connections=connections,
                                                pool_maxsize=max_size,
                                                max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    @property
    def _session(self):
        __session = getattr(self, "__session", None)
        if __session:
            return __session

        session = self.create_session()
        setattr(self, "__session", session)
        return session





from omnisdk.omnitron.endpoints import (ChannelAddressEndpoint,
                                        ChannelCountryEndpoint,
                                        ChannelCityEndpoint,
                                        ChannelTownshipEndpoint,
                                        ChannelDistrictEndpoint)
from omnisdk.omnitron.models import (City, Country, Township, District, Address,
                                     Customer)
from requests import HTTPError

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import AddressDto
from channel_app.omnitron.constants import INTEGRATION_TYPE, ErrorType
from channel_app.omnitron.exceptions import (CountryException,
                                             IntegrationMappingException,
                                             CityException, TownshipException,
                                             DistrictException)


class GetOrCreateAddress(OmnitronCommandInterface):
    endpoint = ChannelAddressEndpoint

    def get_data(self) -> dict:
        """
        :return:
        {
            "email": "john.doe@akinon.com",
            "phone_number": "05556667788",
            "first_name": "John",
            "last_name": "Doe",
            "country": 1,
            "city": 80,
            "line": "Hemen sahil kenarı",
            "title": "COMM-876",
            "township": 933,
            "district": 71387,
            "postcode": "",
            "notes": null,
            "company_name": "",
            "tax_office": "",
            "tax_no": "",
            "e_bill_taxpayer": false,
            "hash_data": "b5e07b50cb033f5e4c86de7f1093bb37",
            "address_type": "customer",
            "retail_store": null,
            "remote_id": null,
            "identity_number": null,
            "extra_field": {},
            "customer": 36406,
            "is_active": true
        },
        """
        address: AddressDto = self.objects["address"]
        customer: Customer = self.objects["customer"]
        country_pk, city_pk, township_pk, district_pk = \
            self.get_location_objects(address.country,
                                      address.city,
                                      address.township,
                                      address.district)
        data = {
            "email": address.email[:254],
            "phone_number": address.phone_number[:128],
            "first_name": customer.first_name[:255],
            "last_name": customer.last_name[:255],
            "customer": customer.pk,
            "country": country_pk,
            "city": city_pk,
            "township": township_pk,
            "district": district_pk,
            "title": address.title and address.title[:128],
            "line": address.line[:255],
            "postcode": address.postcode[:64],
            "notes": address.notes and address.notes[:512],
            "company_name": address.company_name and address.company_name[:255],
            "tax_office": address.tax_office and address.tax_office[:64],
            "tax_no": address.tax_no and address.tax_no[:20],
            "e_bill_taxpayer": address.e_bill_taxpayer or False,
            "remote_id": address.remote_id,
            "identity_number": address.identity_number and address.identity_number[:64],
            "extra_field": address.extra_field or {}
        }
        return data

    def send(self, validated_data) -> object:
        """

        :param validated_data: data for address
        :return: address objects
        """
        try:
            address_obj = Address(**validated_data)
            address = self.endpoint(
                channel_id=self.integration.channel_id).create(
                item=address_obj)
        except HTTPError as e:
            addresses = []
            if ("hash_data" in e.response.text and
                    "non_field_errors" in e.response.text):
                try:
                    error_data = e.response.json()['non_field_errors'].split()[0].split(':')
                    if error_data[0] == "hash_data":
                        hash_data = error_data[1]
                        addresses = self.endpoint(
                            channel_id=self.integration.channel_id).list(
                            params={"hash_data": hash_data})
                except Exception:
                    pass
            if len(addresses) != 1:
                raise
            address = addresses[0]
        return [address]

    def get_location_objects(self, country_code, city_name, township_name=None,
                             district_name=None):
        """
        :param country_code: country code ("Tr")
        :param city_name: city name ("Istanbul")
        :param township_name: town ship ("Kadikoy")
        :param district_name: district name ("Moda")
        :return: country_pk, city_pk, township_pk, district_pk
        """
        country = self.get_country(country_code=country_code)
        country_pk = country.pk

        city = self.get_city(country=country, city_name=city_name)
        city_pk = city.pk
        township = None
        township_pk = None
        if township_name:
            township = self.get_township(country=country, city=city,
                                         township_name=township_name)
            township_pk = township.pk
        district_pk = None
        if district_name and township:
            district = self.get_district(country=country, city=city,
                                         township=township,
                                         district_name=district_name)
            district_pk = district.pk
        return country_pk, city_pk, township_pk, district_pk

    def get_mapping_object(self, code, endpoint):
        """
        :param code: Identifier code for the City and Country on the Sales Channel. This is
            prefix with channel_id as integration_code
        :param endpoint: omnitron sdk endpoint
        :return: mapped objects E.g (City, Country)
        """
        integration_code = f"{self.integration.channel_id}_{code}"
        params = {"mapping__code__exact": integration_code,
                  "mapping__integration_type": INTEGRATION_TYPE}
        objects = endpoint.list(params=params)
        if len(objects) != 1:
            raise IntegrationMappingException(params={"code": integration_code})
        return objects

    def get_country(self, country_code: str) -> Country:
        endpoint = ChannelCountryEndpoint(channel_id=self.integration.channel_id)

        params = {"code__exact": country_code}
        countries = endpoint.list(params=params)
        if len(countries) == 1:
            return countries[0]
        try:
            countries = self.get_mapping_object(country_code, endpoint)
        except IntegrationMappingException as exc:
            raise CountryException(params={
                "type": ErrorType.country.value,
                "remote_code": exc.params['code'],
                "country_code": country_code,
                "name": country_code,
                "integration_type": INTEGRATION_TYPE,
                "err_code": "Country was not found",
                "err_desc": f"Country code {country_code} was not found in Omnitron"})
        return countries[0]

    def get_city(self, country: Country, city_name: str) -> City:
        endpoint = ChannelCityEndpoint(channel_id=self.integration.channel_id)

        params = {"name__exact": city_name, "country": country.pk}
        cities = endpoint.list(params=params)
        if len(cities) == 1:
            return cities[0]
        try:
            cities = self.get_mapping_object(code=city_name, endpoint=endpoint)
        except IntegrationMappingException as exc:
            raise CityException(
                params={"type": ErrorType.city.value,
                        "remote_code": exc.params['code'],
                        "country_id": country.pk,
                        "integration_type": INTEGRATION_TYPE,
                        "err_code": "integrations_200_10_2",
                        "err_desc": f"Address is not valid. channel: {self.integration.channel_id} "
                                    f"code:{exc.params['code']} "
                                    f"model: City country id: {country.pk}"})
        return cities[0]

    def get_township(self, country: Country, city: City, township_name: str) -> Township:
        endpoint = ChannelTownshipEndpoint(channel_id=self.integration.channel_id)

        params = {"name__exact": township_name, "country": country.pk, "city": city.pk}
        townships = endpoint.list(params=params)
        if len(townships) == 1:
            return townships[0]
        try:
            townships = self.get_mapping_object(township_name, endpoint)
        except IntegrationMappingException as exc:
            raise TownshipException(
                params={"type": ErrorType.township.value,
                        "remote_code": exc.params['code'],
                        "city_id": city.pk,
                        "name": township_name,
                        "integration_type": INTEGRATION_TYPE,
                        "err_code": "integrations_200_10_1",
                        "err_desc": f"Address is not valid. channel: {self.integration.channel_id}"
                                    f"code: {exc.params['code']} "
                                    f"model: Township city id: {city.pk}"})
        return townships[0]

    def get_district(self, country: Country, city: City, township: Township,
                     district_name: str) -> District:
        endpoint = ChannelDistrictEndpoint(channel_id=self.integration.channel_id)

        params = {"name__exact": district_name, "country": country.pk, "city": city.pk,
                  "township": township.pk}
        districts = endpoint.list(params=params)
        if len(districts) == 1:
            return districts[0]
        try:
            districts = self.get_mapping_object(district_name, endpoint)
        except IntegrationMappingException as exc:
            raise DistrictException(
                params={"type": ErrorType.district.value,
                        "remote_code": exc.params['code'],
                        "city_id": city.pk,
                        "township_id": township.pk,
                        "name": district_name,
                        "integration_type": INTEGRATION_TYPE,
                        "err_code": "integrations_200_10_3",
                        "err_desc": f"Address is not valid. channel: {self.integration.channel_id}"
                                    f"code: {exc.params['code']} "
                                    f"model: District township id: {township.pk}"}
            )
        return districts[0]



from omnisdk.omnitron.endpoints import ChannelBatchRequestEndpoint
from omnisdk.omnitron.models import BatchRequest

from channel_app.omnitron.constants import BatchRequestStatus


class ClientBatchRequest(object):
    """
    Batch requests work as a state machine. They are used to track state of a flow that must
    be running across multiple systems. It starts at initialized state.
    """
    def __init__(self, channel_id):
        self.channel_id = channel_id
        self.endpoint = ChannelBatchRequestEndpoint()

    def create(self) -> BatchRequest:
        batch_request = BatchRequest(channel=self.channel_id)
        return self.endpoint(channel_id=self.channel_id).create(item=batch_request)

    def to_commit(self, batch_request: BatchRequest) -> BatchRequest:
        """
        Once objects are fetched by Channel app, state is updated to commit and object
        integration are marked(if sent) on Omnitron side.
        :param batch_request:
        :return:
        """
        br = BatchRequest(channel=self.channel_id)
        br.objects = batch_request.objects
        br.status = BatchRequestStatus.commit.value
        br.content_type = batch_request.content_type
        return self.endpoint(channel_id=self.channel_id).update(id=batch_request.pk, item=br)

    def to_sent_to_remote(self, batch_request: BatchRequest) -> BatchRequest:
        """
        If objects are sent to channel, state must be updated to sent_to_remote
        and channel batch id is stored if it sent one.
        :param batch_request:
        :return:
        """
        br = BatchRequest(channel=self.channel_id)
        br.remote_batch_id = batch_request.remote_batch_id
        br.status = BatchRequestStatus.sent_to_remote.value
        return self.endpoint(channel_id=self.channel_id).update(
            id=batch_request.pk, item=br)

    def to_ongoing(self, batch_request: BatchRequest) -> BatchRequest:
        """
        If channel has not finished the batch yet, once we query for it after an interval, we
        update the state to ongoing.
        :param batch_request:
        :return:
        """
        br = BatchRequest(channel=self.channel_id)
        br.status = BatchRequestStatus.ongoing.value
        return self.endpoint(channel_id=self.channel_id).update(
            id=batch_request.pk, item=br)

    def to_fail(self, batch_request: BatchRequest) -> BatchRequest:
        """
        If channel fails completing the batch, batch request is finalized with fail state.
        :param batch_request:
        :return:
        """
        br = BatchRequest(channel=self.channel_id)
        br.objects = batch_request.objects
        br.status = BatchRequestStatus.fail.value
        return self.endpoint(channel_id=self.channel_id).update(
            id=batch_request.pk, item=br)

    def to_done(self, batch_request: BatchRequest) -> BatchRequest:
        """
        If all objects are processed disregarding the fact that they succeeded or failed,
        batch request is finalized with done.
        :param batch_request:
        :return:
        """
        br = BatchRequest(channel=self.channel_id)
        br.objects = batch_request.objects
        br.status = BatchRequestStatus.done.value
        return self.endpoint(channel_id=self.channel_id).update(
            id=batch_request.pk, item=br)




from redis import Redis
from omnisdk.omnitron.client import OmnitronApiClient as BaseOmnitronApiClient


class RedisClient(Redis):
    def __init__(self):
        from channel_app.core import settings
        self.broker_database_index = int(settings.CACHE_DATABASE_INDEX)
        self.broker_port = settings.CACHE_PORT
        self.broker_host = settings.CACHE_HOST
        super(RedisClient, self).__init__(host=self.broker_host,
                                          port=self.broker_port,
                                          db=self.broker_database_index)


class OmnitronApiClient(BaseOmnitronApiClient):
    """
    OmnitronApiClient class for Omnitron API requests.
    """
    client_route = "api/v1/"
    redis_prefix = "omnitron_auth_token_prefix"

    def __init__(self, base_url, username, password):
        self.redis_client = RedisClient()
        super().__init__(base_url, username, password)

    @property
    def token(self):
        token = self.redis_client.get(self.redis_prefix)
        if token:
            token = token.decode("utf-8")
        else:
            # Check redis key for retry count with max 3 attempts in 6 min.
            retry_count = int(self.redis_client.get("retry_count") or 0)
            if retry_count < 3:
                token = self.refresh_key()
                self.redis_client.set("retry_count", retry_count + 1, 360)
            else:
                raise Exception("Login attempts exceeded 3 times in 6 min.")
        return token

    def set_token(self, token):
        self.redis_client.set(self.redis_prefix, token)



from typing import List

from omnisdk.omnitron.models import (ProductStock, Product, IntegrationAction,
                                     BatchRequest)

from channel_app.core import settings
from channel_app.core.data import (ProductBatchRequestResponseDto,
                                   ErrorReportDto)
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import ContentType


class ProductService(object):
    batch_service = ClientBatchRequest

    def insert_products(self, add_mapped=True, add_stock=True, add_price=True,
                        add_categories=True, is_sync=True,
                        is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.product.value) as omnitron_integration:
            products = omnitron_integration.do_action(
                key='get_inserted_products')

            if add_mapped:
                products = products and omnitron_integration.do_action(
                    key='get_mapped_products', objects=products)

            if add_stock:
                products = products and omnitron_integration.do_action(
                    key='get_product_stocks', objects=products)

            if add_price:
                products = products and omnitron_integration.do_action(
                    key='get_product_prices', objects=products)

            if add_categories:
                products = products and omnitron_integration.do_action(
                    key='get_product_categories', objects=products)

            if not products:
                return

            products: List[Product]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_inserted_products',
                objects=products,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[ProductBatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[Product]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_product_batch_requests',
                    objects=response_data)

    def update_products(self, add_mapped=True, add_stock=True, add_price=True,
                        add_categories=True, is_sync=True, is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.product.value) as omnitron_integration:
            products = omnitron_integration.do_action(
                key='get_updated_products')
            if add_mapped:
                products = products and omnitron_integration.do_action(
                    key='get_mapped_products', objects=products)

            if add_stock:
                products = products and omnitron_integration.do_action(
                    key='get_product_stocks', objects=products)

            if add_price:
                products = products and omnitron_integration.do_action(
                    key='get_product_prices', objects=products)

            if add_categories:
                products = products and omnitron_integration.do_action(
                    key='get_product_categories', objects=products)

            if not products:
                return

            products: List[Product]

            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_products',
                objects=products,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[ProductBatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[Product]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_product_batch_requests',
                    objects=response_data)

    def delete_products(self, is_sync=True, is_content_object=True,
                        is_success_log=True):
        with OmnitronIntegration(
                content_type=ContentType.integration_action.value) as omnitron_integration:
            products_integration_action = omnitron_integration.do_action(
                key='get_deleted_products')
            if not products_integration_action:
                return
            products_integration_action = omnitron_integration.do_action(
                key="get_content_objects_from_integrations",
                objects=products_integration_action
            )
            products_integration_action: List[IntegrationAction]
            response_data, reports, data = ChannelIntegration().do_action(
                key='send_deleted_products',
                objects=products_integration_action,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[ProductBatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[IntegrationAction]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_commit(
                        batch_request=omnitron_integration.batch_request)
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_delete_product_batch_requests',
                    objects=response_data)

    def get_delete_product_batch_requests(self, is_success_log=True):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.integration_action.value})
            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, report, data = ChannelIntegration().do_action(
                    key='check_deleted_products',
                    objects=batch_request)

                # tips
                response_data: List[ProductBatchRequestResponseDto]
                report: ErrorReportDto
                data: BatchRequest

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)
                if response_data:
                    omnitron_integration.batch_request = batch_request
                    omnitron_integration.do_action(
                        key='process_delete_product_batch_requests',
                        objects=batch_request)

    def get_product_batch_requests(self, is_success_log=True):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.product.value})
            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, report, data = ChannelIntegration().do_action(
                    key='check_products', objects=batch_request)

                # tips
                response_data: List[ProductBatchRequestResponseDto]
                report: ErrorReportDto
                data: BatchRequest

                if report and (is_success_log or not report.is_ok):
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request
                    omnitron_integration.do_action(
                        key='process_product_batch_requests',
                        objects=response_data)



from omnisdk.omnitron.endpoints import ChannelErrorReportEndpoint, \
    ContentTypeEndpoint
from omnisdk.omnitron.models import ErrorReport

from channel_app.core.commands import OmnitronCommandInterface
from channel_app.core.data import ErrorReportDto
from channel_app.omnitron.constants import CONTENT_TYPE_IDS


class CreateAddressErrorReports(OmnitronCommandInterface):
    endpoint = ChannelErrorReportEndpoint

    def get_data(self) -> list:
        data = self.objects
        return data

    def send(self, validated_data) -> object:
        endpoint = self.endpoint(
            channel_id=self.integration.channel_id,
            raw=True,
            path=validated_data["type"])
        obj = endpoint.create(item=validated_data)
        return obj

    def row_send_error_report(self):
        pass

    def send_error_report(self, raw_request, raw_response):
        pass

    def check_run(self, is_ok, formatted_data):
        return False

class CreateErrorReports(OmnitronCommandInterface):
    endpoint = ChannelErrorReportEndpoint

    def get_data(self) -> object:
        return self.objects

    def validated_data(self, data):
        assert isinstance(data, ErrorReportDto)
        return data

    def send(self, validated_data: ErrorReportDto) -> object:
        error_report = ErrorReport(
            action_content_type=self.get_content_type(
                validated_data.action_content_type),
            action_object_id=validated_data.action_object_id,
            target_content_type=self.get_content_type(
                validated_data.target_content_type or "channel"),
            target_object_id=validated_data.target_object_id or self.integration.channel_id,
            obj_modified_date=validated_data.modified_date,
            error_code=validated_data.error_code or "custom",
            error_desc=validated_data.error_description or "custom",
            raw_request=validated_data.raw_request,
            raw_response=validated_data.raw_response,
            is_ok=validated_data.is_ok
        )
        report = self.endpoint(
            channel_id=self.integration.channel_id).create(item=error_report)
        return [report]

    def get_content_type(self, content_type: str):
        try:
            return CONTENT_TYPE_IDS[content_type]
        except KeyError:
            pass
        content_type = ContentTypeEndpoint().list(
            params={"model": content_type})
        if not content_type:
            raise Exception("Invalid Content Type")
        if len(content_type) > 1:
            raise Exception("Invalid Content Type")

        content_type_id = content_type[0].id
        CONTENT_TYPE_IDS[content_type[0].model] = content_type_id
        return content_type_id

    def row_send_error_report(self):
        pass

    def send_error_report(self, raw_request, raw_response):
        pass

    def check_run(self, is_ok, formatted_data):
        return False



from typing import List

from omnisdk.omnitron.models import ProductStock, BatchRequest

from channel_app.core import settings
from channel_app.core.settings import OmnitronIntegration, ChannelIntegration
from channel_app.core.data import BatchRequestResponseDto, ErrorReportDto
from channel_app.omnitron.batch_request import ClientBatchRequest
from channel_app.omnitron.constants import ContentType


class StockService(object):
    batch_service = ClientBatchRequest

    def update_product_stocks(self, is_sync=True, is_success_log=True,
                              add_product_objects=False):
        with OmnitronIntegration(
                content_type=ContentType.product_stock.value) as omnitron_integration:
            product_stocks = omnitron_integration.do_action(
                key='get_updated_stocks')

            if add_product_objects:
                product_stocks = product_stocks and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_stocks)

            if not product_stocks:
                return

            product_stocks: List[ProductStock]
            response_data, reports, data = ChannelIntegration().do_action(
                key='send_updated_stocks',
                objects=product_stocks,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)

            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductStock]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_stock_batch_requests',
                    objects=response_data)


    def update_product_stocks_from_extra_stock_list(self, is_sync=True,
                                                    is_success_log=True,
                                                    add_product_objects=False):
        warehouse_mappings = self.get_warehouse_mappings()
        for stock_list_id, country_code in warehouse_mappings.items():
            with OmnitronIntegration(
                    content_type=ContentType.product_stock.value) as omnitron_integration:
                product_stocks = omnitron_integration.do_action(
                    key='get_updated_stocks_from_extra_stock_list',
                    object=stock_list_id)

                if add_product_objects:
                    product_stocks = product_stocks and omnitron_integration.do_action(
                        key='get_product_objects', objects=product_stocks)

                if not product_stocks:
                    return

                product_stocks: List[ProductStock]
                response_data, reports, data = ChannelIntegration().do_action(
                    key='send_updated_stocks',
                    objects=(product_stocks, country_code),
                    batch_request=omnitron_integration.batch_request,
                    is_sync=is_sync)

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: List[ProductStock]

                if not is_sync:
                    if reports[0].is_ok:
                        self.batch_service(
                            settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                            batch_request=omnitron_integration.batch_request)
                    else:
                        is_sync = True

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if is_sync:
                    omnitron_integration.do_action(
                        key='process_stock_batch_requests',
                        objects=response_data)

    def insert_product_stocks_from_extra_stock_list(self, is_sync=True,
                                                    is_success_log=True,
                                                    add_product_objects=False):
        warehouse_mappings = self.get_warehouse_mappings()
        for stock_list_id, country_code in warehouse_mappings.items():
            with OmnitronIntegration(
                    content_type=ContentType.product_stock.value) as omnitron_integration:
                product_stocks = omnitron_integration.do_action(
                    key='get_inserted_stocks_from_extra_stock_list',
                    object=stock_list_id)

                if add_product_objects:
                    product_stocks = product_stocks and omnitron_integration.do_action(
                        key='get_product_objects', objects=product_stocks)

                if not product_stocks:
                    return

                product_stocks: List[ProductStock]
                response_data, reports, data = ChannelIntegration().do_action(
                    key='send_inserted_stocks',
                    objects=(product_stocks, country_code),
                    batch_request=omnitron_integration.batch_request,
                    is_sync=is_sync)

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: List[ProductStock]

                if not is_sync:
                    if reports[0].is_ok:
                        self.batch_service(
                            settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                            batch_request=omnitron_integration.batch_request)
                    else:
                        is_sync = True

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if is_sync:
                    omnitron_integration.do_action(
                        key='process_stock_batch_requests',
                        objects=response_data)

    def insert_product_stocks(self, is_sync=True, is_success_log=True,
                              add_product_objects=False):
        with OmnitronIntegration(
                content_type=ContentType.product_stock.value) as omnitron_integration:
            product_stocks = omnitron_integration.do_action(
                key='get_inserted_stocks')

            if add_product_objects:
                product_stocks = product_stocks and omnitron_integration.do_action(
                    key='get_product_objects', objects=product_stocks)

            if not product_stocks:
                return

            product_stocks: List[ProductStock]
            response_data, reports, data = ChannelIntegration().do_action(
                key='send_inserted_stocks',
                objects=product_stocks,
                batch_request=omnitron_integration.batch_request,
                is_sync=is_sync)
            # tips
            response_data: List[BatchRequestResponseDto]
            reports: List[ErrorReportDto]
            data: List[ProductStock]

            if not is_sync:
                if reports[0].is_ok:
                    self.batch_service(
                        settings.OMNITRON_CHANNEL_ID).to_sent_to_remote(
                        batch_request=omnitron_integration.batch_request)
                else:
                    is_sync = True

            if reports and (is_success_log or not reports[0].is_ok):
                for report in reports:
                    omnitron_integration.do_action(
                        key='create_error_report',
                        objects=report)

            if is_sync:
                omnitron_integration.do_action(
                    key='process_stock_batch_requests',
                    objects=response_data)

    def get_stock_batch_requests(self, is_success_log=True):
        with OmnitronIntegration(create_batch=False) as omnitron_integration:
            batch_request_data = omnitron_integration.do_action(
                'get_batch_requests',
                params={
                    "status": ["sent_to_remote", "ongoing"],
                    "content_type": ContentType.product_stock.value})
            # tips
            batch_request_data: List[BatchRequest]

            for batch_request in batch_request_data:
                response_data, reports, data = ChannelIntegration().do_action(
                    key='check_stocks',
                    objects=batch_request,
                    batch_request=batch_request
                )

                # tips
                response_data: List[BatchRequestResponseDto]
                reports: List[ErrorReportDto]
                data: BatchRequest

                if reports and (is_success_log or not reports[0].is_ok):
                    for report in reports:
                        omnitron_integration.do_action(
                            key='create_error_report',
                            objects=report)

                if response_data:
                    omnitron_integration.batch_request = batch_request
                    omnitron_integration.do_action(
                        key='process_stock_batch_requests',
                        objects=response_data)

    def get_warehouse_mappings(self):
        """
        Process is like below:
         We get key value list of the warehouse mapping, filter Omnitron
         with key field
        {
            "1": "ae",
            "2": "us"
        }
        :return:
        """
        with OmnitronIntegration(create_batch=False) as integration:
            warehouse_mapping = integration.channel.conf.get(
                "WAREHOUSE_CODES", {})
            stock_lists = integration.catalog.extra_stock_lists
            if integration.catalog.stock_list:
                stock_lists.append(integration.catalog.stock_list)
            filtered_warehouse_mapping = {
                stock_list: warehouse_mapping[stock_list]
                for stock_list in map(str, stock_lists)
                if stock_list in warehouse_mapping}
        return filtered_warehouse_mapping



from redis import Redis
import unittest
from unittest.mock import patch, Mock

from channel_app.core.clients import OmnitronApiClient
from omnisdk.exceptions import ValidationError


class TestOmnitronApiClient(unittest.TestCase):
    """
    run:
        python -m unittest channel_app.core.tests.TestOmnitronApiClient
    """
    redis_con = Redis()
    omnitron_auth_url = "https://example.com/api/v1/auth/login/"
    test_token = "Token test_key"

    @patch("channel_app.core.clients.RedisClient")
    @patch('requests.Session')
    def setUp(self, mock_session, mock_redis_client):
        self.redis_con.flushall()
        mock_redis_client.return_value = self.redis_con
        # Create a mock session and set it as the return value of
        # requests.Session()
        self.mock_session = mock_session.return_value
        self.mock_session.post.return_value = Mock(
            status_code=200,
            json=Mock(return_value={"key": "test_key"})
        )
        self.base_url = "https://example.com/"
        self.username = "testuser"
        self.password = "testpassword"
        self.client = OmnitronApiClient(
            self.base_url,
            self.username,
            self.password
        )

    def test_token_property(self):
        # Assert that the session.post() method was called once
        self.mock_session.post.assert_called_once_with(
            self.omnitron_auth_url,
            json={"username": "testuser", "password": "testpassword"}
        )
        # Assert that the self.client.redis_prefix key is set to the token
        self.assertEqual(
            self.redis_con.get(self.client.redis_prefix).decode("utf-8"),
            self.test_token
        )
        # Call token property and assert that the token is returned
        for _ in range(2):
            token = self.client.token
            self.assertEqual(
                token,
                self.test_token
            )

        # Assert that the session.post() method was called only once
        self.mock_session.post.assert_called_once_with(
            self.omnitron_auth_url,
            json={"username": "testuser", "password": "testpassword"}
        )

    def test_refresh_key(self):
        self.redis_con.flushall()
        self.mock_session.reset_mock()
        self.mock_session.post.return_value = Mock(
            status_code=200,
            json=Mock(return_value={"key": "updated_key"})
        )
        self.client.refresh_key()
        # Assert that the session.post() method was called once
        self.mock_session.post.assert_called_once_with(
            self.omnitron_auth_url,
            json={"username": "testuser", "password": "testpassword"}
        )
        # Assert that the self.client.redis_prefix key is set to the token
        self.assertEqual(
            self.redis_con.get(self.client.redis_prefix).decode("utf-8"),
            "Token updated_key"
        )

    def test_refresh_key_with_authentication_error(self):
        self.redis_con.flushall()
        self.mock_session.reset_mock()
        self.mock_session.post.return_value = Mock(
            status_code=400,
            json=Mock(return_value={"error_message": "Authentication failed"})
        )

        with self.assertRaises(ValidationError) as context:
            self.client.refresh_key()

        self.assertIn("Authentication failed", str(context.exception))
        self.mock_session.post.assert_called_once_with(
            self.omnitron_auth_url,
            json={"username": "testuser", "password": "testpassword"}
        )
        # Assert that the self.client.redis_prefix key is not set
        self.assertIsNone(
            self.redis_con.get(self.client.redis_prefix)
        )

    def test_set_token(self):
        # Get token from redis which is set at initialization of the client
        token = self.client.token
        self.assertEqual(
            token,
            self.test_token
        )
        new_token = "Token test_new_key"
        self.client.set_token(new_token)
        self.assertEqual(
            self.client.token,
            new_token
        )
        # Assert that the self.client.redis_prefix key is set to the token
        self.assertEqual(
            self.redis_con.get(self.client.redis_prefix).decode("utf-8"),
            new_token
        )


    def test_retry_count_with(self):
        # Get token from redis which is set at initialization of the client
        token = self.client.token
        self.assertEqual(
            token,
            self.test_token
        )
        new_token = "Token test_new_key"
        self.client.set_token(new_token)
        self.assertEqual(
            self.client.token,
            new_token
        )
        # Assert that the self.client.redis_prefix key is set to the token
        self.assertEqual(
            self.redis_con.get(self.client.redis_prefix).decode("utf-8"),
            new_token
        )
        self.assertEqual(
            int(self.redis_con.get("retry_count").decode("utf-8")),
            1
        )

    def test_retry_count(self):
        self.redis_con.flushall()
        for i in range(3):
            token = self.client.token
            self.assertEqual(
                token,
                self.test_token
            )
            self.redis_con.delete(self.client.redis_prefix)
        with self.assertRaises(Exception) as context:
            self.client.token
        self.assertIn(
            "Login attempts exceeded 3 times in 6 min.",
            str(context.exception)
        )
        self.assertEqual(
            int(self.redis_con.get("retry_count").decode("utf-8")),
            i + 1
        )
